#include <cassert>
#include <chrono>
#include <cstring>
#include <memory>
#include <cublas_v2.h>
#include <cudnn.h>
#include <iostream>
#include <sstream>
#include <string.h>
#include <fstream>
#include <time.h>
#include <unordered_map>
#include <vector>
#include <cuda_runtime.h>
#include <cuda.h>


#include "NvInfer.h"
#include "NvInferPlugin.h"
#include "NvUffParser.h"
#include "SplitPlugin.h"
//#include "common.h"

#include <opencv2/opencv.hpp>

using namespace nvinfer1;
using namespace nvuffparser;
using namespace plugin;
using namespace std;



string uffFilename = "../data/split.uff";
string planFilename = "../data/split.plan";
string inputName = "input_image";
int inputHeight = 5;
int inputWidth = 5;
int inputChannel = 3;
string outputName = "pred";
int maxBatchSize = 1;
int maxWorkspaceSize = 0;

int outputHeight = 5;
int outputWidth = 5;
int outputChannel = 3;

class Logger : public ILogger {
public:
    void log(ILogger::Severity severity, const char* msg) override {
        // suppress info-level messages
        if (severity == Severity::kINFO) return;

        switch (severity) {
        case Severity::kINTERNAL_ERROR: std::cerr << "INTERNAL_ERROR: "; break;
        case Severity::kERROR: std::cerr << "ERROR: "; break;
        case Severity::kWARNING: std::cerr << "WARNING: "; break;
        case Severity::kINFO: std::cerr << "INFO: "; break;
        default: std::cerr << "UNKNOWN: "; break;
        }
        std::cerr << msg << std::endl;
    }
} gLogger;



template <class T>
void hwc2chw(T* src, T* dst, int h, int w, int c) {
  assert(src != 0);
  assert(dst != 0);
  for (int j=0; j<c; j++) {
    T* p = src + j;
    for (int i=0; i<h; i++) {
      for (int k=0; k<w; k++) {
        *dst = *p;
        ++dst;
        p += c;
      }
    }
  }
}

ICudaEngine* loadModelAndCreateEngine(const char* uffFile, int maxBatchSize,
                                      int maxWorkspaceSize, IUffParser* parser, 
                                      nvuffparser::IPluginFactory* pluginFactory, IHostMemory*& trtModelStream)
{
    // Create the builder
    IBuilder* builder = createInferBuilder(gLogger);

    // Parse the UFF model to populate the network, then set the outputs.
    INetworkDefinition* network = builder->createNetwork();
    parser->setPluginFactory(pluginFactory);

    std::cout << "Begin parsing model..." << std::endl;
    //if (!parser->parse(uffFile, *network, nvinfer1::DataType::kFLOAT))
    //    RETURN_AND_LOG(nullptr, ERROR, "Fail to parse");

    if (!parser->parse(uffFilename.c_str(), *network, nvinfer1::DataType::kFLOAT))
    {
      cout << "Failed to parse UFF\n";
    }
    std::cout << "End parsing model..." << std::endl;

    // Build the engine.
    builder->setMaxBatchSize(maxBatchSize);
    builder->setMaxWorkspaceSize(1 << 20);
    builder->setHalf2Mode(false);

    std::cout << "Begin building engine..." << std::endl;
    ICudaEngine* engine = builder->buildCudaEngine(*network);
    //if (!engine)
    //    RETURN_AND_LOG(nullptr, ERROR, "Unable to create engine");
    std::cout << "End building engine..." << std::endl;

    // We don't need the network any more, and we can destroy the parser.
    network->destroy();
    parser->destroy();

    // Serialize the engine, then close everything down.
    trtModelStream = engine->serialize();

    builder->destroy();
    shutdownProtobufLibrary();
    return engine;
}


void doInference(IExecutionContext& context, char* input_buffer, int input_size, char* output_buffer, int output_size, int batchSize){

  assert(input_size == inputHeight * inputWidth * inputChannel * sizeof(float) );
  assert(output_size == outputHeight * outputWidth * outputChannel * sizeof(float) );

  const ICudaEngine& engine = context.getEngine();
  // Input and output buffer pointers that we pass to the engine - the engine requires exactly IEngine::getNbBindings(),
  // of these, but in this case we know that there is exactly 1 input and 2 output.
  //int nbBindings = engine.getNbBindings();

  auto t_start = std::chrono::high_resolution_clock::now();

  // input and output buffer pointers that we pass to the engine - the engine requires exactly IEngine::getNbBindings(),
	// of these, but in this case we know that there is exactly one input and one output.
	assert(engine.getNbBindings() == 2);
	void* buffers[2];

	// In order to bind the buffers, we need to know the names of the input and output tensors.
	// note that indices are guaranteed to be less than IEngine::getNbBindings()

	int inputIndex = 0; 
	int	outputIndex = engine.getNbBindings() - 1;


	// create GPU buffers and a stream
	CHECK(cudaMalloc(&buffers[inputIndex],  input_size ));
	CHECK(cudaMalloc(&buffers[outputIndex], output_size ));

	cudaStream_t stream;
	CHECK(cudaStreamCreate(&stream));

	// DMA the input to the GPU,  execute the batch asynchronously, and DMA it back:
	CHECK(cudaMemcpyAsync(buffers[inputIndex], input_buffer, input_size , cudaMemcpyHostToDevice, stream));
	context.enqueue(batchSize, buffers, stream, nullptr);
	CHECK(cudaMemcpyAsync(output_buffer, buffers[outputIndex], output_size, cudaMemcpyDeviceToHost, stream));
	cudaStreamSynchronize(stream);

	// release the stream and the buffers
	cudaStreamDestroy(stream);
	CHECK(cudaFree(buffers[inputIndex]));
	CHECK(cudaFree(buffers[outputIndex]));

  auto t_end = std::chrono::high_resolution_clock::now();
  auto ms = std::chrono::duration<float, std::milli>(t_end - t_start).count();
	std::cout << "Network inference time: " << ms << " ms." << std::endl;
}



int main(int argc, char* argv[])
{
    IUffParser *parser = createUffParser();
    parser->registerInput(inputName.c_str(), DimsCHW(inputChannel, inputHeight, inputWidth),UffInputOrder::kNCHW);
    parser->registerOutput(outputName.c_str());

    IHostMemory* trtModelStream{nullptr};

    PluginFactory pluginFactorySerialize;

    ICudaEngine* tmpEngine = loadModelAndCreateEngine(uffFilename.c_str(), maxBatchSize, maxWorkspaceSize, parser, &pluginFactorySerialize, trtModelStream);
    assert(tmpEngine != nullptr);
    assert(trtModelStream != nullptr);

    /* serialize engine and write to file */
    ofstream planFile;
    planFile.open(planFilename);
    //IHostMemory *serializedEngine = engine->serialize();
    planFile.write((char *)trtModelStream->data(), trtModelStream->size());
    planFile.close(); 

    tmpEngine->destroy();
    pluginFactorySerialize.destroyPlugin();



    // Available images.
    int input_size = inputHeight * inputWidth * inputChannel * sizeof(float);
    int input_num = inputHeight * inputWidth * inputChannel;
    vector<float> input(input_num);

    vector<float> tf_input(input_num);
    for (int i = 0; i < input_num; i++){
      tf_input[i] = i;
    }
    hwc2chw(&tf_input[0], &input[0], inputHeight, inputWidth,inputChannel) ;

    for (int i =0; i < input_num; i++){
      std::cout << "input_chw "<< i << " : "<< input[i]<< std::endl;
    }

    // Deserialize the engine.
    std::cout << "*** deserializing" << std::endl;
    IRuntime* runtime = createInferRuntime(gLogger);
    assert(runtime != nullptr);
    PluginFactory pluginFactory;
    ICudaEngine* engine = runtime->deserializeCudaEngine(trtModelStream->data(), trtModelStream->size(), &pluginFactory);
    assert(engine != nullptr);
    trtModelStream->destroy();
    IExecutionContext* context = engine->createExecutionContext();

    assert(context != nullptr);

    // allocate memory on host / device for input / output
    int output_size = outputHeight * outputWidth * outputChannel * sizeof(float);
    int output_num = outputHeight * outputWidth * outputChannel;
    vector<float> output(output_num);

 
    // Run inference.
    int batchSize = 1;
    doInference(*context, (char *)&input[0], input_size, (char *)&output[0], output_size, batchSize);
  
    for (int i =0; i < output_num; i ++ ){
      std::cout << "output "<< i << " : "<< output[i]<< std::endl;
     }

    // Destroy the engine.
    context->destroy();
    engine->destroy();
    runtime->destroy();

    // Destroy plugins created by factory
    pluginFactory.destroyPlugin();

    return 0;
}
